/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BANKING_BANK_H_
#define DRN_BANKING_BANK_H_

#include <iosfwd>
#include <map>

#include <QString>

#include <accounting/AccountNumber.h>
#include <banking_export.h>

#include "BankName.h"
#include "BankAccountType.h"


class QDebug;

namespace drn::banking
{

/**
 * @brief A representation of a financial institution.
 */
class DRN_BANKING_EXPORT Bank
{
	std::map<accounting::AccountNumber, BankAccountTypes> accountNumbers_;
	bool isClosed_{};

public:

	Bank() = default;
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception BankError Thrown if the provided name is empty. This ensures that every financial
	 * institution always has a valid and meaningful name.
	 * @exception BankError When the account number is invalid.
	 * @exception BankError When the account type is unknown.
	 *
	 * @param name The name to assign to the financial institution. Must represent a valid name.
	 * @param numbers A collection of account numbers to associate with the financial institution.
	 * Each account number must map to a valid bank account type.
	 * @param isClosed Indicates whether the financial institution should be marked as closed for
	 * usage. Pass true to initialise the financial institution as closed; otherwise, false.
	 */
	explicit Bank(
		BankName name,
		std::map<accounting::AccountNumber, BankAccountTypes> numbers = {},
		const bool isClosed = false
	);
	/**
	 * @brief An accessor for all the accounts associated with this financial institution.
	 *
	 * @return The account numbers of the associated accounts, empty when not set.
	 */
	[[nodiscard]]
	const std::map<accounting::AccountNumber, BankAccountTypes>& accounts() const noexcept;
	/**
	* @brief An accessor for the flag to indicate if the financial institution is closed for usage.
	*
	* @return Gives true when the financial institution is closed.
	*/
	[[nodiscard]]
	bool isClosed() const noexcept;
	/**
	 * @brief Marks the financial institution as closed for usage.
	 *
	 * @exception BankError When the financial institution is already closed.
	 */
	void close();
	/**
	 * @brief Marks the financial institution as open for usage.
	 *
	 * @exception BankError When the financial institution is already opened.
	 */
	void open();
	/**
	 * @brief Adds a new account number along with its corresponding type to the financial
	 * institution.
	 *
	 * It is designed to register an account number and its associated type to the list of accounts
	 * in the financial institution.
	 *
	 * @exception BankError When the account number is invalid.
	 * @exception BankError When the account type is unknown.
	 *
	 * @param number The account number being added to the financial institution.
	 * @param type The account type corresponding to the account number.
	 *
	 * @post The account number and its type will be registered in the system.
	 */
	void add(const accounting::AccountNumber& number, const BankAccountTypes type);
	/**
	 * @brief Removes an account number from the association with the financial institution.
	 *
	 * @exception BankError When the account number to remove is not associated with the financial
	 * institution.
	 *
	 * @param number The number to remove.
	 */
	void remove(const accounting::AccountNumber& number);

	/// The name of the financial institution, empty when not set.
	BankName name_;
};

/**
 * @brief Full equality comparison operator.
 *
 * @param lhs The left-hand-side to compare against
 * @param rhs The right-hand-side to compare with
 *
 * @return Gives true when all values are the same, false otherwise.
 */
DRN_BANKING_EXPORT bool operator==(const Bank& lhs, const Bank& rhs);
/**
 * @brief Full inequality comparison operator.
 *
 * @param lhs The left-hand-side to compare against
 * @param rhs The right-hand-side to compare with
 *
 * @return Gives true when any value is /not/ the same, false otherwise.
 */
DRN_BANKING_EXPORT bool operator!=(const Bank& lhs, const Bank& rhs);
/**
 * @brief Full object stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param b The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT std::ostream& operator<<(std::ostream& out, const Bank& b);
/**
 * @brief Full object Qt stream insertion operator.
 *
 * @param out The stream to place the data into.
 * @param b The data to place into the stream.
 *
 * @return The modified stream after insertion.
 */
DRN_BANKING_EXPORT ::QDebug operator<<(::QDebug out, const Bank& b);

}

#endif
